from typing import Dict, Optional, Tuple, Union

from dict_zip import dict_zip, dict_zip_longest

d1: Dict[str, int] = {"a": 1, "b": 2, "c": 3}
d2: Dict[str, int] = {"a": 4, "b": 5}
d3: Dict[str, int] = {"b": 6, "c": 7}
d4: Dict[str, float] = {"b": 6.0, "c": 7.0}
d5: Dict[str, float] = {"b": 6.0, "c": 7.0, "d": 8.0}

t1 = dict_zip(d1, d2)
t1_: Dict[str, Tuple[int, int]] = t1

t2 = dict_zip(d1, d2, d3)
t2_: Dict[str, Tuple[int, int, int]] = t2

t3 = dict_zip(d3, d4)
t3_: Dict[str, Tuple[int, float]] = t3

t4 = dict_zip(d1, d2, d3, d4)
t4_: Dict[str, Tuple[int, int, int, float]] = t4

t5 = dict_zip_longest(d1, d2)
t5_: Dict[str, Tuple[Optional[int], Optional[int]]] = t5

t6 = dict_zip_longest(d1, d2, fillvalue=4)
t6_: Dict[str, Tuple[int, int]] = t6

t7 = dict_zip_longest(d1, d4)
t7_: Dict[str, Tuple[Optional[int], Optional[float]]] = t7

t8 = dict_zip_longest(d1, d4, fillvalue=4)
t8_: Dict[str, Tuple[int, Union[float, int]]] = t8

t9 = dict_zip_longest(d1, d4, fillvalue=4.0)
t9_: Dict[str, Tuple[Union[int, float], float]] = t9



"""dict_zip

This module provides a function that concatenates dictionaries.
Like the zip function for lists, it concatenates dictionaries.

Example:
    >>> from dict_zip import dict_zip
    >>> dict_zip({'a': 1, 'b': 2}, {'a': 3, 'b': 4})
    {'a': (1, 3), 'b': (2, 4)}

    >>> from dict_zip import dict_zip_longest
    >>> dict_zip_longest({'a': 1, 'b': 2, 'c': 4}, {'a': 3, 'b': 4})
    {'a': (1, 3), 'b': (2, 4), 'c': (4, None)}
"""

import functools


def dict_zip(*dictionaries):
    """Returns a new dictionary \
concatenated with the dictionaries specified in the argument.

    The key is a common key that each dictionary has.
    The value is a tuple of the values of the dictionaries.

    >>> dict_zip({'a': 1, 'b': 2}, {'a': 3, 'b': 4})
    {'a': (1, 3), 'b': (2, 4)}
    """
    common_keys = functools.reduce(
        lambda x, y: x & y, (set(d.keys()) for d in dictionaries)
    )
    return_dic = {}
    for dic in dictionaries:
        for key, val in dic.items():
            if key in common_keys:
                return_dic.setdefault(key, []).append(val)

    return {key: tuple(val) for key, val in return_dic.items()}


def dict_zip_longest(*dictionaries, fillvalue=None):
    """Returns a new dictionary \
concatenated with the dictionaries specified in the argument.

    The keys are the union set of the dictionaries.
    The value is a tuple of the values of the dictionaries.
    If the specified dictionary does not have the key, \
it is filled with fillvalue (default: None).

    >>> dict_zip_longest({'a': 1, 'b': 2, 'c': 4}, {'a': 3, 'b': 4})
    {'a': (1, 3), 'b': (2, 4), 'c': (4, None)}
    """
    all_keys = __all_keys(d.keys() for d in dictionaries)
    return_dic = {key: tuple() for key in all_keys}

    for dic in dictionaries:
        for key in all_keys:
            return_dic[key] += (dic.get(key, fillvalue),)

    return return_dic


def __all_keys(iterables):
    keys = []
    for iterable in iterables:
        for key in iterable:
            if key in keys:
                continue
            keys.append(key)
    return keys
